# Stubs for mersad.test.util.test_base_class (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest
from typing import Any

class TestMersadClassicalBase(unittest.TestCase):
    BaseClass: Any = ...
    def setUp(self) -> None: ...
    def test_config(self) -> None: ...
    def test_config_reset(self) -> None: ...
    def test_process_replace_manual_key(self) -> None: ...
    def test_return_key(self) -> None: ...
    def test_config_bad_type(self) -> None: ...

class TestMainFunctionClassical(unittest.TestCase):
    main: Any = ...
    def setUp(self) -> None: ...
    def test_parent_parser(self) -> None: ...



# Stubs for mersad.test.classical (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Names in __all__ with no definition:
#   test_affine_cipher
#   test_shift_cipher



# Stubs for mersad.util.string_analyzer (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
from typing import Dict
from typing import List
from typing import Set

def find_unique_letters(text: str) -> Set[str]: ...
def find_letter_indexes(text: str, letter: str) -> List[int]: ...
def map_letters_to_indexes(text: str) -> Dict[str, List[int]]: ...



# mersad/test/classical/__init__.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.test package.
====================

This package contains tests
for classical package of Mersad library.

"""

# Python Standard Library
from typing import List

# please keep alphabetical order
__all__: List[str] = [
    "test_affine_cipher",
    "test_shift_cipher"
]



# Stubs for mersad.util.base_class (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import argparse
from typing import Any
from typing import List
from typing import Optional
from typing import Union

class MersadClassicalBase:
    _defaults: Any = ...
    configuration: Any = ...
    def __init__(self, **kwargs: Union[int, str, bool]) -> None: ...
    def __str__(self) -> str: ...
    def encrypt(self, plain_text: str, key: Optional[int]=..., replace_key: bool=...) -> str: ...
    def decrypt(self, cipher_text: str, key: Optional[int]=..., replace_key: bool=...) -> str: ...
    def config(self, **kwargs: Union[int, str, bool]) -> None: ...
    def reset(self) -> None: ...
    def show_key(self) -> int: ...
    def _process(self, text: str, key: Optional[int], replace_key: bool, decrypt: bool) -> str: ...
    @staticmethod
    def _translator(text: str, **kwargs: Union[int, str, bool]) -> str: ...

class MainFunctionClassical:
    args: Any = ...
    agent_class: Any = ...
    description: Any = ...
    epilog: Any = ...
    def __init__(self, args: List[Any], agent_class: Any, description: str, epilog: str) -> None: ...
    def process(self) -> None: ...
    def parse_args(self) -> argparse.Namespace: ...
    @staticmethod
    def _argparse_parent() -> argparse.ArgumentParser: ...



# Stubs for mersad.test.util.test_type_check (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest

class TestTypeCheck(unittest.TestCase):
    def test_different_types(self) -> None: ...



# Stubs for mersad.test.util.test_crypto_math (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest

class TestCryptoMath(unittest.TestCase):
    def test_extended_gcd(self) -> None: ...
    def test_mod_inverse(self) -> None: ...



# Stubs for mersad.util (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Names in __all__ with no definition:
#   base_class
#   crypto_math
#   string_analyzer
#   string_manipulation
#   type_check



# mersad/util/base_class.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.util.base_class module.
==============================

The module contains base (parent) classes for
Mersad's cipher classes.

"""

# Python Standard Library
import argparse
import string
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# Mersad Library
from mersad._version import __version__
from mersad.util import type_check


class MersadClassicalBase(object):
    """
    Base class for classical cipher algorithms objects.

    This class provides a clean interface for building other
    cipher classes with it.

    Sub-classes should override self._translator method
    with their desired algorithm to make their class work
    as they want.
    """

    def __init__(self, **kwargs: Union[int, str, bool]) -> None:
        r"""
        Create an instance of the class.

        Example
        ======

        >>> agent = MersadClassicalBase(key=53, letter_sequence="abcdefghrstuvwxyz")

        AffineCipher takes keyword arguments when initializing new instance.

        Valid kwargs
        ============
        key             : (optional) primary key.
        letter_sequence : (optional) alphabet for using in cipher.
        shuffle         : (optional) randomize letter sequence order.
        seed            : (optional)(requires shuffle) specifies a seed for
                          randomizing, default seed is 0.

        Agent uses predefined default values for each of above arguments if
        it isn't provided by the user, for above example the key and letter sequence
        are provided by user but not shuffle and seed, agent will use default values
        for shuffle and seed.

        Default key is set to 0.
        Default letter sequence is set to "string.printable" except \r character.
        Default shuffle is set to False.
        Default seed is set to 0.

        Each instance has it's own unique configurations saved in
        self.configuration and can work independent from other instances.

        Configurations can be changed even after initialization via
        self.config() method.

        Example
        =======

        >>> agent = MersadClassicalBase()
        >>> # override defaults
        >>> agent.config(key=3453, letter_sequence="abcdefghijklmnopqrstuvwxyz")

        Note
        ====

        This class is just a parent class and can't encrypt/decrypt anything.
        """
        # private configuration dictionary that holds default values.
        # should not be changed by anyone!
        self._defaults: Dict[str, Any] = dict(
                key=0, letter_sequence=string.printable.replace("\r", ""),
                shuffle=False, seed=0, decrypt=False
        )
        # public configuration dictionary.
        self.configuration: Dict[str, Any] = dict()
        # set self.configuration to default values (default_configuration).
        self.reset()
        # process kwargs and update self.configuration values.
        self.config(**kwargs)

    def __str__(self) -> str:
        """
        Return the objects info as string.

        :return : information such as key, letter sequence and etc.
        :rtype  : str
        """
        key: int = self.configuration["key"]
        letter_sequence: str = self.configuration["letter_sequence"]
        shuffle: bool = self.configuration["shuffle"]
        seed: int = self.configuration["seed"]
        return "{0}: {1}.\n{2}: {3}.\n{4}: {5}.\n{6}: {7}.\n{8}: {9}.".format(
                "Agent type", type(self), "key", key, "letter_sequence",
                letter_sequence, "shuffle", shuffle, "seed", seed
        )

    def encrypt(self, plain_text: str, key: Optional[int] = None,
                replace_key: bool = False) -> str:
        """
        Encrypt a string.

        This function is a wrapper for self._process function.

        :param plain_text   :   (required) the string that will be encrypted.
        :param key          :   (optional) a new key for encryption.
        :param replace_key  :   (optional) if set to True, the provided new key
                                replaces old key at self.configuration.
        :return             :   encrypted string.
        :rtype              :   str
        """
        return self._process(plain_text, key, replace_key=replace_key, decrypt=False)

    def decrypt(self, cipher_text: str, key: Optional[int] = None,
                replace_key: bool = False) -> str:
        """
        Decrypt a string.

        This function is a wrapper for self._process function.

        :param cipher_text  :   (required) the string that will be decrypted.
        :param key          :   (optional) a new key for decryption.
        :param replace_key  :   (optional) if set to True, the provided new key
                                replaces old key at self.configuration.
        :return             :   decrypted string.
        :rtype              :   str
        """
        return self._process(cipher_text, key, replace_key=replace_key, decrypt=True)

    def config(self, **kwargs: Union[int, str, bool]) -> None:
        """
        Assign values to self.configuration dictionary.

        :raise ValueError: if type of a dictionary value is wrong.
        """
        if "letter_sequence" in kwargs:
            type_check.type_guard(kwargs["letter_sequence"], str)
            self.configuration["letter_sequence"] = kwargs["letter_sequence"]

        if "key" in kwargs and kwargs["key"] is not None:
            type_check.type_guard(kwargs["key"], int)
            self.configuration["key"] = kwargs["key"]

        if "seed" in kwargs:
            type_check.type_guard(kwargs["seed"], int)
            self.configuration["seed"] = kwargs["seed"]

        if "shuffle" in kwargs:
            type_check.type_guard(kwargs["shuffle"], bool)
            self.configuration["shuffle"] = kwargs["shuffle"]

        if "decrypt" in kwargs:
            type_check.type_guard(kwargs["decrypt"], bool)
            self.configuration["decrypt"] = kwargs["decrypt"]

    def reset(self) -> None:
        """Reset all configurations to defaults."""
        # deep copy default_configuration dictionary
        # into instance variable self.configuration.
        self.configuration = {i: j for (i, j) in self._defaults.items()}

    def show_key(self) -> int:
        """
        Return the current key in self.configuration dictionary.

        :return : current key in use.
        :rtype  : int
        """
        return self.configuration["key"]

    def _process(self, text: str, key: Optional[int], replace_key: bool,
                 decrypt: bool) -> str:
        """
        Handle the process for both encryption and decryption.

        :param text         : string to be processed.
        :param key          : key for encryption/decryption.
        :param decrypt      : switch for encryption/decryption.
        :param replace_key  : replace the old key in self.configuration
                              with new one.
        :return             : encrypted/decrypted string.
        :rtype              : str
        """
        # explicitly switch mode to encryption/decryption.
        self.config(decrypt=decrypt)

        # replace the old key in self.configuration with new key
        if replace_key:
            self.config(key=key)

        if key and not replace_key:
            # deep copy self.configuration dictionary into new dictionary to be used.
            configuration = {i: j for (i, j) in self.configuration.items()}
            # check key type to be compatible.
            type_check.type_guard(key, int)
            # change the key in the copied dictionary,
            # self.configuration key will remain unchanged.
            configuration["key"] = key
        else:
            # alias self.configuration dictionary to configuration
            # in this case configuration dictionary isn't a copy
            # and is just a pointer to self.configuration
            configuration = self.configuration

        # return a call to affine_cipher_translator function with
        # string and unpacked (prepended "**" is used to unpack the dictionary)
        # configuration dictionary as arguments.
        return self._translator(text, **configuration)

    @staticmethod
    def _translator(text: str, **kwargs: Union[int, str, bool]) -> str:
        """
        Wrap the actual encryption/decryption function for class.

        This method should be implemented in sub classes.

        :param text : string to be translated.
        :return     : translated text.
        :rtype      : str
        """
        return "Not Implemented in this class!"


class MainFunctionClassical(object):
    """
    Manage classical cipher programs execution from terminal.

    This class has a parser and a process function that controls
    all thw procedures needed for running classical ciphers from terminal.
    """

    def __init__(self, args: List[Any], agent_class, description: str, epilog: str) \
            -> None:
        """Initialize instance with needed data."""
        self.args: List[Any] = args
        self.agent_class = agent_class
        self.description: str = description
        self.epilog: str = epilog

    def process(self) -> None:
        """Process program execution based on terminal arguments."""
        args: argparse.Namespace = self.parse_args()
        # load text_input from file or terminal.
        # type annotations.
        text_input: str
        if args.file:
            with open(args.file, "r") as file:
                text_input = file.read()
        else:
            text_input = args.text

        # construct a shift cipher agent with parsed arguments.
        agent = self.agent_class(
                key=args.key,
                letter_sequence=args.letters,
                shuffle=args.shuffle,
                seed=args.seed
        )
        # type annotations.
        text_output: str
        if args.decrypt:
            text_output = agent.decrypt(text_input)
        else:
            text_output = agent.encrypt(text_input)

        # write output to a file or show on terminal.
        if args.output:
            with open(args.output, "w+") as file:
                file.write(text_output)
        else:
            print(text_output)

    def parse_args(self) -> argparse.Namespace:
        """
        Start parsing terminal arguments.

        :return: terminal argument namespace.
        :rtype: argparse.Namespace
        """
        # create a parser with given arguments
        parser: argparse.ArgumentParser = argparse.ArgumentParser(
                parents=[self._argparse_parent()],
                description=self.description,
                epilog=self.epilog,
                formatter_class=argparse.RawDescriptionHelpFormatter
        )

        # parse args and create a dictionary like namespace object.
        return parser.parse_args(args=self.args)

    @staticmethod
    def _argparse_parent() -> argparse.ArgumentParser:
        """
        Create base parser object.

        This function creates and returns a base parser
        for all classical ciphers.

        :return: parser
        :rtype: argparse.ArgumentParser
        """
        # create the parent parser, the base parser for creating
        # various other program specific parsers at the top of it.
        parent_parser: argparse.ArgumentParser = argparse.ArgumentParser(
                add_help=False
        )

        # display version
        version: str = f"Azadeh Afzar - Mersad Cryptography Library v{__version__}"
        parent_parser.add_argument("-V", "--version", action="version",
                                   version=version)

        # create an mutually exclusive group for parser, user should either
        # provide a filename or a text for the process.
        source_type = parent_parser.add_mutually_exclusive_group(required=True)

        help_file: str = "file path for reading data from it"
        source_type.add_argument("-f", "--file", type=str, help=help_file)

        help_text: str = "read data from terminal"
        source_type.add_argument("-t", "--text", type=str, help=help_text)

        help_output: str = "file path for writing the result into it"
        parent_parser.add_argument("-o", "--output", type=str, help=help_output)

        help_key: str = "key for encryption/decryption"
        parent_parser.add_argument("-k", "--key", type=int, help=help_key)

        help_decrypt: str = "decrypt data"
        parent_parser.add_argument("-d", "--decrypt", action="store_true",
                                   default=False, help=help_decrypt)

        help_letters: str = "alphabet for encryption/decryption"
        parent_parser.add_argument("-l", "--letters", type=str,
                                   default=string.printable, help=help_letters)

        help_shuffle: str = "shuffle alphabet letters"
        parent_parser.add_argument("-sh", "--shuffle", action="store_true",
                                   default=False, help=help_shuffle)

        help_seed: str = "specify random seed for shuffling the alphabet letters"
        parent_parser.add_argument("-s", "--seed", type=int, default=0,
                                   help=help_seed)

        return parent_parser



# Stubs for mersad.classical.shift_cipher (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
from typing import Union

# Mersad Library
from mersad.util.base_class import MersadClassicalBase

def main() -> None: ...

class ShiftCipher(MersadClassicalBase):
    @staticmethod
    def _translator(text: str, **kwargs: Union[int, str, bool]) -> str: ...

def shift_cipher_translator(text: str, **kwargs: Union[int, str, bool]) -> str: ...



#!/usr/bin/env python3

# mersad/classical/affine_cipher.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.classical.affine_cipher module.
======================================

The affine is a type of mono alphabetical substitution
cipher, wherein each letter in an alphabet is mapped to
its numeric equivalent, encrypted using a single mathematical
function, and converted back to a letter.

"""

# Python Standard Library
import sys
from math import gcd
from typing import Dict
from typing import Union

# Mersad Library
from mersad.util import crypto_math
from mersad.util import string_manipulation
from mersad.util.base_class import MainFunctionClassical
from mersad.util.base_class import MersadClassicalBase


def main() -> None:
    """Execute program in terminal (cli application)."""
    # module descriptions.
    description: str = "Azadeh Afzar - Mersad Affine Cipher\n" \
                       + "Encrypt/Decrypt data with Affine"
    epilog: str = "CIA can still read your messages ..."

    # create a parser and parse command line arguments.
    program = MainFunctionClassical(sys.argv[1:], AffineCipher, description, epilog)
    program.process()


class AffineCipher(MersadClassicalBase):
    """
    Affine cipher algorithm class.

    Create an instance like every other python class.

    Example:
    ==================================

    >>> agent = AffineCipher(key=235, letter_sequence="abcdefghijklmnopqrstuvwxyz")
    >>> # encrypt a string
    >>> agent.encrypt("Is this really more secure than shift cipher?")
    'Ih qmvh ylbwwj fxyl hltzyl qmbo hmvuq tvgmly?'

    ==================================

    AffineCipher takes keyword arguments when initializing new instance.

    Valid kwargs keys are listed in below:
        - key             : (optional) primary key.
        - letter_sequence : (optional) alphabet for using in cipher.
        - shuffle         : (optional) randomize letter sequence order.
        - seed            : (optional)(requires shuffle) specifies a seed
                            for randomizing, default seed is 0.

    Agent uses predefined default values for each of above arguments if
    it isn't provided by the user, for above example the key and letter sequence
    are provided by user but not shuffle and seed, agent will use default values
    for shuffle and seed.

    Default key is set to 0.
    Default letter sequence is set to "string.printable" except "\r".
    Default shuffle is set to False.
    Default seed is set to 0.

    Each instance has it's own unique configurations saved in
    self.configuration and can work independent from other instances.

    Configurations can be changed even after initialization via
    self.config() method.

    Example:
    ==================================

    >>> # create object without any keyword arguments so agent will use
    >>> # default values for all the settings.
    >>> agent = AffineCipher()
    >>> # override defaults.
    >>> agent.config(key=135)
    >>> # encrypt a string
    >>> agent.encrypt("Is this really more secure than shift cipher?")
    '<"t#QR"t!NJUU(tVX!Nt"NL$!Nt#QJWt"QRO#tLRYQN!h'

    ==================================

    Warning:
    always provide a key to when initializing AffineCipher, because the default key
    will raise ValueError.
    """

    @staticmethod
    def _translator(text: str, **kwargs: Union[int, str, bool]) -> str:
        """
        Wrap the actual encryption/decryption function for class.

        :param text : string to be translated.
        :return     : translated text
        :rtype      : str
        """
        return affine_cipher_translator(text, **kwargs)


def affine_cipher_translator(text: str, **kwargs: Union[int, str, bool]) -> str:
    """
    Translate a string with Affine cipher algorithm.

    :param text                             : string to be translated.
    :param kwargs:
        key                                 : key for encrypt/decrypt.
        letter_sequence                     : the letter sequence which will be
                                              used for shifting letters.
        shuffle (optional)(default = False) : randomize letter sequence order.
        seed (optional)(requires shuffle)   : specify a seed for randomizing,
                                              default seed is 0.
        decrypt (optional)(default = False) : switch for encryption/decryption mode.
    :return                                 : translated text
    :rtype                                  : str
    """
    # for sake of readability and prettifying below code
    # I will assign aliases for key, values inside kwargs.
    sequence: str = kwargs["letter_sequence"]
    # length of sequence is needed for mathematical calculations.
    sequence_length: int = len(sequence)
    # key alias.
    key: int = kwargs["key"]
    # type annotations
    key_a: int
    key_b: int
    # generate partial keys.
    key_a, key_b = divmod(key, sequence_length)
    # validate keys.
    _check_keys(key_a, key_b, sequence_length)
    # default shuffle to False if no shuffle is defined in kwargs.
    shuffle: bool = kwargs["shuffle"] if "shuffle" in kwargs else False
    # default seed to 0 if no seed is defined in kwargs.
    seed: int = kwargs["seed"] if "seed" in kwargs else 0
    # default decrypt to False if no decrypt is defined in kwargs.
    decrypt: bool = kwargs["decrypt"] if "decrypt" in kwargs else False

    # type annotations
    translated_sequence: Dict[str, str]
    # blank string.
    translated: str = ""

    # shuffle letter sequence with respect to seed if shuffle is set to True.
    if shuffle:
        # shuffle sequence
        sequence = string_manipulation.shuffle_string(sequence, seed)

    # create a table mapping that maps every letter in sequence to
    # it's equivalent new letter with respect to the key and size of sequence.
    if decrypt:
        # find mode inverse of key a.
        key_a_mode_inverse = crypto_math.mod_inverse(key_a, sequence_length)
        # generate decryption letter sequence map.
        translated_sequence = {
            i: sequence[
                ((sequence.index(i) - key_b) * key_a_mode_inverse) % sequence_length
                ]
            for i in sequence
        }
    else:
        # generate encryption letter sequence map.
        translated_sequence = {
            i: sequence[(sequence.index(i) * key_a + key_b) % sequence_length]
            for i in sequence
        }

    # select each letter in the text and only if it is also provided in sequence
    # from user and replace it with new letter selected by affine method.
    for letter in text:
        if letter in sequence:
            # get the translated letter for this letter from mapping.
            translated_letter = translated_sequence[letter]
        else:
            # if the letter in the text isn't in sequence, it remains unchanged.
            translated_letter = letter
        # add new letter to translated string.
        translated += translated_letter

    return translated


def _check_keys(key_a: int, key_b: int, sequence_length: int) -> None:
    """
    Check if keys are in correct range.

    :param key_a            : first key.
    :param key_b            : second key.
    :param sequence_length  : length of letter sequence.
    :raise ValueError       : if keys are incorrect.
    """
    # test for invalid key range
    if key_a <= 0:
        raise ValueError(
                "The affine cipher's 'key a' must be greater than 0. "
                + "Change your key."
        )

    if not (0 <= key_b <= sequence_length - 1):
        raise ValueError(
                "The affine cipher's 'key b' must be in this range: "
                + "0 <= 'key b' <= length of letter sequence - 1. "
                + "Change your key."
        )

    if gcd(key_a, sequence_length) != 1:
        print(sequence_length)
        print(key_a)
        raise ValueError(
                "The affine cipher's 'key a' and length of 'letter "
                + "sequence' are not relatively prime. Change your key."
        )



# Stubs for mersad.util.crypto_math (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
from typing import Tuple

def extended_gcd(a: int, b: int) -> Tuple[int, int, int]: ...
def mod_inverse(a: int, b: int) -> int: ...



# Stubs for mersad (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Mersad Library
from mersad import _version as _version

# Names in __all__ with no definition:
#   classical
#   test
#   util



# mersad/util/string_manipulation.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.util.string_manipulation module.
=======================================

This module provides functions for modifying string objects.

"""

# Python Standard Library
import random
from typing import List


def shuffle_string(text: str, seed: int) -> str:
    """
    Shuffle letters in the string with random.shuffle.

    with seed parameter in can create identical strings
    from same source in multiple calls.

    :param text : string to be shuffled.
    :param seed : number for setting seed on random library.
    :return     : the shuffled string.
    :rtype      : str
    """
    random.seed(seed)
    letter_list: List[str] = list(text)
    random.shuffle(letter_list)
    return ''.join(letter_list)


def replace_letter_by_index(text: str, letter: str, indexes: List[int]) -> str:
    """
    Replace letters in specific index of a string with another letter.

    :param text     : string to be manipulated.
    :param letter   : the new letter which will be placed instead of old letters.
    :param indexes  : a list of indexes in the string that should be replaced with
                      new letter.
    :return         : new string with replaced letters.
    :rtype          : str
    """
    letter_list: List[str] = list(text)
    for index in indexes:
        letter_list[index] = letter
    return ''.join(letter_list)



#!/usr/bin/env python3

# mersad/classical/shift_cipher.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.classical.shift_cipher module.
=====================================

The shift (also knows as caesar cipher) is a type
of mono alphabetical substitution cipher, in which
each letter in alphabet is replaced by a letter some
fixed number of positions down the alphabet.

"""

# Python Standard Library
import sys
from typing import Dict
from typing import Union

# Mersad Library
from mersad.util import string_manipulation
from mersad.util.base_class import MainFunctionClassical
from mersad.util.base_class import MersadClassicalBase


def main() -> None:
    """Execute program in terminal (cli application)."""
    # module descriptions.
    description: str = "Azadeh Afzar - Mersad Shift Cipher\n" \
                       + "Encrypt/Decrypt data with Shift algorithm"
    epilog: str = "Oh you think it is a safe way to hide your secrets from NSA?"

    # create a parser and parse command line arguments.
    program = MainFunctionClassical(sys.argv[1:], ShiftCipher, description, epilog)
    program.process()


class ShiftCipher(MersadClassicalBase):
    """
    Shift cipher algorithm class.

    Create an instance like every other python class.

    Example:
    ==================================

    >>> agent = ShiftCipher(key=3453, letter_sequence="abcdefghijklmnopqrstuvwxyz")
    >>> # encrypt a string
    >>> agent.encrypt("Hail Julius Caesar.")
    "Hvdg Jpgdpn Cvznvm."

    ==================================

    ShiftCipher takes keyword arguments when initializing new instance.

    Valid kwargs keys are listed in below:
        - key             : (optional) primary key.
        - letter_sequence : (optional) alphabet for using in cipher.
        - shuffle         : (optional) randomize letter sequence order.
        - seed            : (optional)(requires shuffle) specifies a seed
                            for randomizing, default seed is 0.

    Agent uses predefined default values for each of above arguments if
    it isn't provided by the user, for above example the key and letter sequence
    are provided by user but not shuffle and seed, agent will use default values
    for shuffle and seed.

    Default key is set to 0.
    Default letter sequence is set to "string.printable" except "\r".
    Default shuffle is set to False.
    Default seed is set to 0.

    Each instance has it's own unique configurations saved in
    self.configuration and can work independent from other instances.

    Configurations can be changed even after initialization via
    self.config() method.

    Example:
    ==================================

    >>> # create object without any keyword arguments so agent will use
    >>> # default values for all the settings.
    >>> agent = ShiftCipher()
    >>> # override defaults.
    >>> agent.config(key=3453, letter_sequence="abcdefghijklmnopqrstuvwxyz")
    >>> # encrypt a string
    >>> agent.encrypt("Hail Julius Caesar.")
    "Hvdg Jpgdpn Cvznvm."

    ==================================
    """

    @staticmethod
    def _translator(text: str, **kwargs: Union[int, str, bool]) -> str:
        """
        Wrap the actual encryption/decryption function for class.

        :param text : string to be translated.
        :return     : translated text
        :rtype      : str
        """
        return shift_cipher_translator(text, **kwargs)


def shift_cipher_translator(text: str, **kwargs: Union[int, str, bool]) -> str:
    """
    Translate a string with Shift cipher algorithm.

    Shift encryption/decryption algorithm:
    1. choose a sequence of letters.
    2. calculate its length.
    3. select letters of a string one by one.
    4. if the selected letter is also in the sequence.
    - 4.1. if selected letter isn't in sequence, it remains unchanged.
    5. find the letter index in the sequence.
    6. add key to index (if decrypting: subtract key from index).
    7. find the reminder of dividing the number you have obtained from step #6
       to sequence length.
    8. find new letter in the sequence at index of the number you have obtained
       from step #7.
    9. replace old letter with new letter.

    :param text                             : string to be translated.
    :param kwargs:
        key                                 : key for encrypt/decrypt.
        letter_sequence                     : the letter sequence which will be
                                              used for shifting letters.
        shuffle (optional)(default = False) : randomize letter sequence order.
        seed (optional)(requires shuffle)   : specify a seed for randomizing,
                                              default seed is 0.
        decrypt (optional)(default = False) : switch for encryption/decryption mode.
    :return                                 : translated text
    :rtype                                  : str
    """
    # for sake of readability and prettifying below code
    # I will assign aliases for key, values inside kwargs.
    sequence: str = kwargs["letter_sequence"]
    key: int = kwargs["key"]
    # length of sequence is needed for mathematical calculations.
    key_size: int = len(sequence)
    # default shuffle to False if no shuffle is defined in kwargs.
    shuffle: bool = kwargs["shuffle"] if "shuffle" in kwargs else False
    # default seed to 0 if no seed is defined in kwargs.
    seed: int = kwargs["seed"] if "seed" in kwargs else 0
    # default decrypt to False if no decrypt is defined in kwargs.
    decrypt: bool = kwargs["decrypt"] if "decrypt" in kwargs else False

    # type annotations
    translated_sequence: Dict[str, str]
    # blank string.
    translated: str = ""

    # shuffle letter sequence with respect to seed if shuffle is set to True.
    if shuffle:
        # shuffle sequence.
        sequence = string_manipulation.shuffle_string(sequence, seed)

    # reverse key for decryption.
    if decrypt:
        key *= -1

    # create a table mapping that maps every letter in sequence to
    # it's equivalent new letter with respect to the key and size of sequence.
    translated_sequence = {
        i: sequence[(sequence.index(i) + key) % key_size] for i in sequence
    }

    # type annotations
    translated_letter: str

    # select each letter in the text and only if it is also provided in sequence
    # from user and replace it with new letter selected by shift method.
    for letter in text:
        if letter in sequence:
            # below code is steps 6 to 8 all together which is actually executed
            # in initializing the translated_sequence dictionary.
            # get the translated letter for this letter from mapping
            translated_letter = translated_sequence[letter]
        else:
            # if the letter in the text isn't in sequence, it remains unchanged.
            translated_letter = letter
        # step 9
        # add new letter to translated string.
        translated += translated_letter

    return translated



# mersad/util/string_analyzer.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#


"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.util.string_analyzer module.
===================================

This module provides functions to extract data
about a string.

"""

# Python Standard Library
from typing import Dict
from typing import List
from typing import Set

# Mersad Library
from mersad.util import type_check


def find_unique_letters(text: str) -> Set[str]:
    """
    Create a set of all letters in a string.

    :param text: string source.
    :return: a set of all unique letters in the string
    :rtype: set
    """
    # check types
    type_check.type_guard(text, str)
    # create and return set
    return set(text)


def find_letter_indexes(text: str, letter: str) -> List[int]:
    """
    List indexes of a letter in a string.

    This function returns a list of index of every occurrence
    of a letter in a string.

    :param text: string source
    :param letter: the letter to find all of its  occurrence indexes
    :return: list containing letters occurrence indexes
    :rtype: list
    """
    # check types
    type_check.type_guard(text, str)
    type_check.type_guard(letter, str)
    # return index list
    return [
        index for index, character in enumerate(text) if character == letter
    ]


def map_letters_to_indexes(text: str) -> Dict[str, List[int]]:
    """
    Create an index  mapping dictionary.

    Dictionary holds that maps every letter
    to a list of indexes of its occurrence in a given string

    :param text: string source for finding indexes letters in it.
    :return: a dictionary with letter as key and a list of indexes as value.
    :rtype: dict
    """
    # check types
    type_check.type_guard(text, str)

    # return a dictionary with letter as key and a list of indexes as value
    return {
        letter: find_letter_indexes(text, letter)
        for letter in find_unique_letters(text)
    }



# mersad/test/classical/test_atbash_cipher.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

# Python Standard Library
import os
import string
import unittest

# 3rd Party Library
from ErfanIO import ReaderIO

# Mersad Library
from mersad.classical.atbash_cipher import AtbashCipher


class TestShiftCipher(unittest.TestCase):
    def setUp(self) -> None:
        # setup path
        classical_path = os.path.abspath(os.path.dirname(__file__))
        test_path = os.path.abspath(os.path.dirname(classical_path))
        self.base_path = os.path.join(test_path, "asset", "texts")
        # create a cipher agent
        self.agent = AtbashCipher()
        self.plain_text = ReaderIO.read(os.path.join(self.base_path, "Long License File.txt"), "text")
        self.sh0_s0 = ReaderIO.read(os.path.join(self.base_path, "AtbashCipher-LLF-sh0-s0.txt"), "text")
        self.sh1_s0 = ReaderIO.read(os.path.join(self.base_path, "AtbashCipher-LLF-sh1-s0.txt"), "text")
        self.custom_alphabet = ReaderIO.read(os.path.join(self.base_path, "AtbashCipher-LLF-alphabet-ascii-lowercase-sh0-s0.txt"), "text")

    def test_encrypt_without_shuffle(self):
        self.agent.config(key=25, shuffle=False, seed=0)
        self.assertEqual(self.sh0_s0, self.agent.encrypt(self.plain_text))

    def test_encrypt_with_shuffle_without_seed(self):
        self.agent.config(key=173, shuffle=True, seed=0)
        self.assertEqual(self.sh1_s0, self.agent.encrypt(self.plain_text))

    def test_encrypt_with_custom_alphabet(self):
        alphabet = string.ascii_lowercase
        self.agent.config(key=85, letter_sequence=alphabet, shuffle=False, seed=0)
        self.assertEqual(self.custom_alphabet, self.agent.encrypt(self.plain_text))

    def test_decryption_without_shuffle(self):
        self.agent.config(key=25, shuffle=False, seed=0)
        self.assertEqual(self.plain_text, self.agent.decrypt(self.sh0_s0))

    def test_decrypt_with_shuffle_without_seed(self):
        self.agent.config(key=173, shuffle=True, seed=0)
        self.assertEqual(self.plain_text, self.agent.decrypt(self.sh1_s0))

    def test_decrypt_with_custom_alphabet(self):
        alphabet = string.ascii_lowercase
        self.agent.config(key=85, letter_sequence=alphabet, shuffle=False, seed=0)
        self.assertEqual(self.plain_text, self.agent.decrypt(self.custom_alphabet))


if __name__ == '__main__':
    unittest.main()



# mersad/test/util/test_base_class.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

# Python Standard Library
import string
import unittest

# Mersad Library
from mersad.util.base_class import MainFunctionClassical
from mersad.util.base_class import MersadClassicalBase


class TestMersadClassicalBase(unittest.TestCase):
    def setUp(self) -> None:
        self.BaseClass = MersadClassicalBase()

    def test_config(self):
        self.BaseClass.config(key=12, shuffle=True, seed=23, decrypt=True)
        self.assertEqual(12, self.BaseClass.configuration["key"])
        self.assertEqual(string.printable.replace("\r", ""),
                         self.BaseClass.configuration["letter_sequence"])
        self.assertEqual(23, self.BaseClass.configuration["seed"])
        self.assertEqual(True, self.BaseClass.configuration["shuffle"])
        self.assertEqual(True, self.BaseClass.configuration["decrypt"])

    def test_config_reset(self):
        self.BaseClass.config(shuffle=True, seed=23, decrypt=True)
        self.BaseClass.reset()
        self.assertEqual(0, self.BaseClass.configuration["key"])
        self.assertEqual(string.printable.replace("\r", ""),
                         self.BaseClass.configuration["letter_sequence"])
        self.assertEqual(False, self.BaseClass.configuration["shuffle"])
        self.assertEqual(0, self.BaseClass.configuration["seed"])
        self.assertEqual(False, self.BaseClass.configuration["decrypt"])

    def test_process_replace_manual_key(self):
        self.BaseClass.config(key=12)
        self.BaseClass.encrypt("It's a dummy text!", key=16, replace_key=True)
        self.assertEqual(16, self.BaseClass.configuration["key"])

    def test_return_key(self):
        self.BaseClass.config(key=12)
        self.assertEqual(12, self.BaseClass.show_key())

    def test_config_bad_type(self):
        with self.assertRaises(TypeError):
            self.BaseClass.config(letter_sequence=12)

        with self.assertRaises(TypeError):
            self.BaseClass.config(key="Hello There!")

        with self.assertRaises(TypeError):
            self.BaseClass.config(seed=True)

        with self.assertRaises(TypeError):
            self.BaseClass.config(shuffle=2)

        with self.assertRaises(TypeError):
            self.BaseClass.config(decrypt="False")


class TestMainFunctionClassical(unittest.TestCase):
    def setUp(self) -> None:
        args = ["10", "--text", "Yup, not funny"]
        agent_class = MersadClassicalBase
        description = "Dummy Dummy Test."
        epilog = "Oh shit here we go again ..."
        self.main = MainFunctionClassical(args, agent_class, description, epilog)

    def test_parent_parser(self):
        args = ["--key", "23", "--text", "wow, is this a test?", "--output",
                "fail.txt", "--decrypt", "--shuffle", "--seed", "12"]
        parent = self.main._argparse_parent()
        args = parent.parse_args(args)
        self.assertEqual(23, args.key)
        self.assertEqual("wow, is this a test?", args.text)
        self.assertEqual("fail.txt", args.output)
        self.assertEqual(True, args.decrypt)
        self.assertEqual(True, args.shuffle)
        self.assertEqual(12, args.seed)


if __name__ == '__main__':
    unittest.main()



# Stubs for mersad.util.string_manipulation (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
from typing import List

def shuffle_string(text: str, seed: int) -> str: ...
def replace_letter_by_index(text: str, letter: str, indexes: List[int]) -> str: ...



# Stubs for mersad.test.classical.test_shift_cipher (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest
from typing import Any

class TestShiftCipher(unittest.TestCase):
    base_path: Any = ...
    agent: Any = ...
    plain_text: Any = ...
    k25_sh0_s0: Any = ...
    k173_sh1_s0: Any = ...
    custom_alphabet: Any = ...
    def setUp(self) -> None: ...
    def test_encrypt_without_shuffle(self) -> None: ...
    def test_encrypt_with_shuffle_without_seed(self) -> None: ...
    def test_encrypt_with_custom_alphabet(self) -> None: ...
    def test_decryption_without_shuffle(self) -> None: ...
    def test_decrypt_with_shuffle_without_seed(self) -> None: ...
    def test_decrypt_with_custom_alphabet(self) -> None: ...
    def test_temporary_key(self) -> None: ...
    def test_temporary_key_to_permanent(self) -> None: ...



# mersad/test/util/test_string_analyzer.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

# Python Standard Library
import unittest

# Mersad Library
from mersad.util import string_analyzer


class TestFindUniqeLetters(unittest.TestCase):
    def test(self):
        unique_letters = string_analyzer.find_unique_letters(
                "Is it UniQue? Or NoT ?! huh?"
        )
        expected_result = {
            '!', '?', 't', 's', 'h', 'Q',
            'N', 'u', 'U', 'O', 'T', 'i',
            'n', 'e', 'o', 'I', ' ', 'r'
        }
        self.assertEqual(expected_result, unique_letters)


class TestFindLetterIndexes(unittest.TestCase):
    def test(self):
        test_index = string_analyzer.find_letter_indexes(
                "Enumerate is an amazing Python function.", "a"
        )
        expected_result = [6, 13, 16, 18]
        self.assertEqual(expected_result, test_index)


class TestMapLettersToIndexes(unittest.TestCase):
    def test(self):
        test_indexes = string_analyzer.map_letters_to_indexes(
                "Hi Martin, I think my TextAnalyzer is working!"
        )
        expected_result = {'!': [45], 'z': [31], 'o': [39], 'h': [14], 'y': [20, 30],
                           'w': [38], ' ': [2, 10, 12, 18, 21, 34, 37], 's': [36],
                           'i': [1, 7, 15, 35, 42], 'm': [19], 'A': [26], 'H': [0],
                           't': [6, 13, 25], 'T': [22], 'I': [11], 'a': [4, 28],
                           'k': [17, 41], 'e': [23, 32], 'n': [8, 16, 27, 43],
                           'g': [44], 'x': [24], 'r': [5, 33, 40], ',': [9],
                           'M': [3], 'l': [29]}
        self.assertEqual(expected_result, test_indexes)


if __name__ == '__main__':
    unittest.main()



# mersad/test/util/test_crypto_math.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#


# Python Standard Library
import unittest

# Mersad Library
from mersad.util.crypto_math import extended_gcd
from mersad.util.crypto_math import mod_inverse


class TestCryptoMath(unittest.TestCase):
    def test_extended_gcd(self):
        test_extended_gcd_1 = extended_gcd(12134, 15)
        expected_result_1 = (1, -1, 809)
        self.assertEqual(expected_result_1, test_extended_gcd_1)
        test_extended_gcd_2 = extended_gcd(80, 1234)
        expected_result_2 = (2, 108, -7)
        self.assertEqual(expected_result_2, test_extended_gcd_2)

    def test_mod_inverse(self):
        test_mod_inverse_1 = mod_inverse(3, 26)
        expected_result_1 = 9
        self.assertEqual(expected_result_1, test_mod_inverse_1)
        test_mod_inverse_2 = mod_inverse(15413, 26)
        expected_result_2 = 5
        self.assertEqual(expected_result_2, test_mod_inverse_2)


if __name__ == '__main__':
    unittest.main()



# Stubs for mersad.test.util.test_string_manipulation (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest

class TestStringManipulation(unittest.TestCase):
    def test_shuffle_string(self) -> None: ...
    def test_replace_letter_by_index(self) -> None: ...



# mersad/util/crypto_math.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.util.crypto_math module.
===============================

This module provides mathematical functions for
evaluating cipher algorithms.

"""

# Python Standard Library
from typing import Tuple


def extended_gcd(a: int, b: int) -> Tuple[int, int, int]:
    """
    Implement extended Euclidean algorithm for gcd.

    in arithmetic and computer programming, the
    extended Euclidean algorithm is an extension
    to the Euclidean algorithm, and computes, in
    addition to the greatest common divisor of
    integers a and b, also the coefficients of
    Bezout's identity, which are integers x and y
    such that ax+by=gcd(a,b).

    :param a: first number.
    :param b: second number.
    :return: (g, x, y) such that a*x + b*y = g = gcd(a, b)
    :rtype: tuple
    """
    # assign variables
    last_x: int = 0
    last_y: int = 1
    x: int = 1
    y: int = 0
    last_remainder: int = abs(b)
    remainder: int = abs(a)

    while remainder:
        last_remainder, (quotient, remainder) = remainder, divmod(last_remainder,
                                                                  remainder)
        x, last_x = last_x - quotient * x, x
        y, last_y = last_y - quotient * y, y

    # fix signs
    last_x *= (-1 if a < 0 else 1)
    last_y *= (-1 if b < 0 else 1)

    # return (g, x, y) such that a*x + b*y = g = gcd(a, b)
    return last_remainder, last_x, last_y


def mod_inverse(a: int, b: int) -> int:
    """
    Find mod inverse of a mod b.

    :raise ValueError: if a and b are not co-prime.

    :param a: number that we want to find its mod inverse.
    :param b: number which is the mod.
    :return: mod inverse of a mod b
    :rtype: int
    """
    # calculate extended gcd and get (g, x, y) such that a*x + b*y = g = gcd(a, b)
    results: Tuple[int, int, int] = extended_gcd(a, b)
    # unpack g and x
    g: int = results[0]
    x: int = results[1]
    if g != 1:
        raise ValueError("mersad.util.crypto_math Error: a and b must be co-prime.")
    return x % b



# Stubs for mersad.test.util.test_string_analyzer (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
import unittest

class TestFindUniqeLetters(unittest.TestCase):
    def test(self) -> None: ...

class TestFindLetterIndexes(unittest.TestCase):
    def test(self) -> None: ...

class TestMapLettersToIndexes(unittest.TestCase):
    def test(self) -> None: ...



#!/usr/bin/env python3

# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

# Python Standard Library
import os

# 3rd Party Library
from setuptools import find_packages
from setuptools import setup

# get path to this file's directory
base_path = os.path.abspath(os.path.dirname(__file__))

# open _version file
with open(os.path.join(base_path, "mersad", "_version.py")) as file:
    version_file = file.readlines()

# set version to None, so if we didn't find
# a version in _version.py, we can throw an error
version = None

# find version
for line in version_file:
    if "__version_info__: Tuple[int, int, int] = " in line:
        # find tuple inside _version.py and reformat it to
        # standard x.y.z version format
        tuple_left = line.index("(")
        tuple_right = line.index(")")
        version = line[tuple_left + 1:tuple_right].replace(",", ".").replace(" ", "")

# throe error if version not found
if not version:
    raise ValueError("ERROR: version not found at _version.py.")

# open mersad/__init__.py to retrieve package info
with open(os.path.join(base_path, "mersad", "__init__.py"), "r") as file:
    info_file = file.readlines()


# define extractor function
def extract_info(source, name):
    for line in source:
        # find the metadata
        if "__{name}__: str =".format(name=name) in line:
            # strip from left until ", " is also excluded
            line = line[line.index('"') + 1:]
            # return string until "
            return line[:line.index('"')]


# general information about package
name = "mersad"
url = "https://gitlab.com/Azadeh-Afzar/Cryptography/Mersad-Cryptography-Library"
author = extract_info(info_file, "author")
author_email = extract_info(info_file, "email")
license_name = extract_info(info_file, "license")
description = extract_info(info_file, "description")
long_description = open("README.md", "r", encoding="utf-8").read()
classifiers = [
    "Development Status :: 1 - Planning",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "License :: OSI Approved :: zlib/libpng License",
    "Operating System :: Unix",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS :: MacOS X",
    "Environment :: Console",
    "Topic :: Security :: Cryptography",
    "Topic :: Software Development",
    "Topic :: Utilities",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]

# dictionary of packages
packages = find_packages()

# add stub files to package data
package_data = {
    "mersad": [
        "py.typed", "*.pyi"
    ],
    "mersad.classical": [
        "*.pyi"
    ],
    "mersad.test": [
        "*.pyi"
    ],
    "mersad.test.asset": [
        "*.pyi"
    ],
    "mersad.test.asset.texts": [
        "*.pyi"
    ],
    "mersad.test.classical": [
        "*.pyi"
    ],
    "mersad.test.util": [
        "*.pyi"
    ],
    "mersad.util": [
        "*.pyi"
    ]
}

# dependencies
install_requires = ["ErfanIO"]

# command line programs
entry_points = {
    "console_scripts": [
        "mclAffine = mersad.classical.affine_cipher:main",
        "mclAtbash = mersad.classical.atbash_cipher:main",
        "mclShift = mersad.classical.shift_cipher:main"
    ]
}

setup(
        name=name,
        version=version,
        packages=packages,
        package_data=package_data,
        entry_points=entry_points,
        url=url,
        license=license_name,
        author=author,
        author_email=author_email,
        description=description,
        long_description=long_description,
        long_description_content_type="text/markdown",
        platforms="Posix; Windows; MacOS X",
        classifiers=classifiers,
        python_requires=">=3.6",
        install_requires=install_requires
)



# mersad/util/type_check.py
#
# This file is a part of:
# Azadeh Afzar - Mersad Cryptography Library in Python language (AA-MCLpy).
#
# Copyright (C) 2019 Azadeh Afzar
# Copyright (C) 2019 Mohammad Mahdi Baghbani Pourvahid
#
# GNU AFFERO GENERAL PUBLIC LICENSE
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ZLIB LICENSE
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
# claim that you wrote the original software. If you use this software
# in a product, an acknowledgement in the product documentation would be
# appreciated but is not required.
#
# 2. Altered source versions must be plainly marked as such, and must not be
# misrepresented as being the original software.
#
# 3. This notice may not be removed or altered from any source distribution.
#

"""
Azadeh Afzar - Mersad Cryptography Library.

mersad.util.type_check module.
==============================
This module is a runtime type checker for Mersad.

"""


def type_guard(given_object: object, expected_type: type) -> None:
    """
    Compare type of an object with expected type.

    :param given_object: object which will be tested.
    :param expected_type: object's expected type.

    :raise TypeError: if object isn't an instance
                      of the expected type
    """
    if type(given_object) != expected_type:
        raise TypeError



# Stubs for mersad._version (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Python Standard Library
from typing import Tuple

__version_info__: Tuple[int, int, int]
__version__: str
